# This file need to be send to the cluster via .addPyFile to handle the pickle problem
# This is outside the optimus folder on purpose because it cause problem importing optimus when using de udf.
# This can not import any optimus file unless it's imported via addPyFile
import datetime
import math
import os
import re
from ast import literal_eval

import fastnumbers
import pandas as pd
import pendulum
from dask import distributed
from dask.dataframe.core import DataFrame as DaskDataFrame

# This function return True or False if a string can be converted to any datatype.
from optimus.helpers.constants import ProfilerDataTypes, CURRENCIES


def is_datetime_str(_value):
    try:
        pendulum.parse(_value, strict=False)
        return True
    except ValueError:
        return False


def str_to_date_format(_value, date_format):
    # Check this https://stackoverflow.com/questions/17134716/convert-dataframe-column-type-from-string-to-datetime-dd-mm-yyyy-format
    try:
        pendulum.from_format(_value, date_format)
        return True
    except ValueError:
        return False


def str_to_null(_value):
    _value = _value.lower()
    if _value == "null":
        return True
    else:
        return False


def is_null(_value):
    return pd.isnull(_value)


def str_to_data_type(_value, _dtypes):
    """
    Check if value can be parsed to a tuple or and list.
    Because Spark can handle tuples we will try to transform tuples to arrays
    :param _value:
    :param _dtypes:
    :return:
    """
    # return True if isinstance(_value, str) else False
    try:
        if isinstance(literal_eval((_value.encode('ascii', 'ignore')).decode("utf-8")), _dtypes):
            return True
    except (ValueError, SyntaxError, AttributeError):
        return False


def is_list_value(value):
    """
    Check if an object is a list
    :param value:
    :return:
    """
    return isinstance(value, list)


def is_list_str(_value):
    return False
    # return str_to_data_type(_value, (list, tuple))


def is_object_str(_value):
    return False
    # return str_to_data_type(_value, (dict, set))


regex_str = r"."

regex_int = r"(^\d+\.[0]+$)|(^\d+$)"  # For cudf 0.14 regex_int = r"^\d+$" # For cudf 0.14
regex_decimal = r"(^\d+\.\d+$)|(^\d+$)"
regex_non_int_decimal = r"^(\d+\.\d+)$"

regex_boolean = r"true|false|0|1"
regex_boolean_compiled = re.compile(regex_boolean)


def is_bool_str(value, compile=False):
    return str_to(str(value), regex_boolean, regex_boolean_compiled, compile)


regex_gender = r"\bmale\b|\bfemale\b"
regex_gender_compiled = re.compile(regex_gender)


def is_gender(value, compile=False):
    return str_to(value, regex_gender, regex_gender_compiled, compile)


regex_full_url = r"^(?:([A-Za-z0-9]*)\://)?(?:(\w+)\.)?([A-Za-z0-9\-\.]+):?(\d+)?(?:/|$)?(\w+\.?\w+)?\??(\w.+)?$"

regex_url = r"(http|https|ftp|s3):\/\/.?[a-zA-Z]*.\w*.[a-zA-Z0-9]*\/?[a-zA-z_-]*.?[a-zA-Z]*\/?"

regex_url_compiled = re.compile(regex_url, re.IGNORECASE)


def str_to_url(value, compile=False):
    return str_to(value, regex_url, regex_url_compiled, compile)


regex_ip = r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}"
regex_ip_compiled = re.compile(regex_ip, re.IGNORECASE)


def str_to_ip(value, compile=False):
    return str_to(value, regex_ip, regex_ip_compiled, compile)


# regex_email = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)" # This do not work in CUDF/RE2
regex_email = r"^[^@]+@[^@]+\.[a-zA-Z]{2,}$"
regex_email_compiled = re.compile(regex_email, re.IGNORECASE)


def is_email(value, compile=False):
    return str_to(value, regex_email, regex_email_compiled, compile)


# Reference https://www.regular-expressions.info/creditcard.html
# https://codereview.stackexchange.com/questions/74797/credit-card-checking
regex_credit_card_number = (r'(4(?:\d{12}|\d{15})'  # Visa
                            r'|5[1-5]\d{14}'  # Mastercard
                            r'|6011\d{12}'  # Discover (incomplete?)
                            r'|7\d{15}'  # What's this?
                            r'|3[47]\d{13}'  # American Express
                            r')$')

regex_credit_card_compiled = re.compile(regex_credit_card_number)


def is_credit_card_number(value, compile=False):
    return str_to(value, regex_credit_card_number, regex_credit_card_compiled, compile)


regex_zip_code = r"^(\d{4,5}(?:[- ]\d{4})?)$"
regex_zip_code_compiled = re.compile(regex_zip_code, re.IGNORECASE)


def is_zip_code(value, compile=False):
    return str_to(value, regex_zip_code, regex_zip_code_compiled, compile)


def is_missing(value):
    return value == ""


regex_social_security_number = "^([1-9])(?!\1{2}-\1{2}-\1{4})[1-9]{2}-[1-9]{2}-[1-9]{4}"
regex_social_security_number_compiled = re.compile(regex_social_security_number, re.IGNORECASE)


def is_social_security_number(value, compile=False):
    return str_to(value, regex_social_security_number, regex_social_security_number_compiled, compile)


regex_http_code = "/^[1-5][0-9][0-9]$/"
regex_http_code_compiled = re.compile(regex_http_code, re.IGNORECASE)


def is_http_code(value, compile=False):
    return str_to(value, regex_http_code, regex_http_code_compiled, compile)


# Reference https://stackoverflow.com/questions/8634139/phone-validation-regex
regex_phone_number = r"\+?\(?([0-9]{3})\)?[-.]?\(?([0-9]{3})\)?[-.]?\(?([0-9]{4})\)?"
regex_phone_number_compiled = re.compile(regex_phone_number, re.IGNORECASE)


def is_phone_number(value, compile=False):
    # print("value",value, str_to(value, regex_phone_number, regex_phone_number_compiled, compile))
    return str_to(value, regex_phone_number, regex_phone_number_compiled, compile)


def str_to(value, regex, compiled_regex, compile=False):
    value = str(value)
    if value is None:
        result = False
    else:
        if compile is True:
            regex = compiled_regex
        else:
            regex = regex

        result = bool(re.match(regex, value))
    return result


def str_to_int(_value):
    return True if fastnumbers.isint(_value) else False


def str_to_decimal(_value):
    return True if fastnumbers.isfloat(_value) else False


def str_to_str(_value):
    return True if isinstance(_value, str) else False


regex_currencies = "|".join(list(CURRENCIES.keys()))
regex_currencies_compiled = re.compile(regex_currencies)


def str_to_currency(value, compile=False):
    return str_to(value, regex_boolean, regex_boolean_compiled, compile)


def is_nan(value):
    """
    Check if a value is nan
    :param value:
    :return:
    """
    result = False
    if is_str(value):
        if value.lower() == "nan":
            result = True
    elif math.isnan(value):
        result = True
    return result


def is_none(value):
    """
    Check if a value is none
    :param value:
    :return:
    """
    result = False
    if is_str(value):
        if value.lower() == "none":
            result = True
    elif value is None:
        result = True
    return result


def is_same_class(class1, class2):
    """
    Check if 2 class are the same
    :param class1:
    :param class2:
    :return:
    """
    return class1 == class2


def is_(value, type_):
    """
    Check if a value is instance of a class
    :param value:
    :param type_:
    :return:
    """

    return isinstance(value, type_)


def is_type(type1, type2):
    """
    Check if a value is a specific class
    :param type1:
    :param type2:
    :return:
    """
    return type1 == type2


def is_function(value):
    """
    Check if a param is a function
    :param value: object to check for
    :return:
    """
    return hasattr(value, '__call__')


def is_list(value, mode=None):
    """
    Check if a string or any not string value is a python list
    :param value:
    :return:
    """
    if mode == "string":
        result = is_list_str(value)
    else:
        result = is_list_value(value)

    return result


def is_object(value):
    """
    Check if a string or any not string value is a python list
    :param value:
    :return:
    """
    return True if is_object_value(value) or is_object_str(value) else False


def is_bool(value):
    if is_bool_value(value):
        return True
    elif is_bool_value(value):
        return True
    return False


def is_list_empty(value):
    """
    Check is a list is empty
    :param value:
    :return:
    """
    return len(value) == 0


def is_dict(value):
    """
    Check if an object is a list
    :param value:
    :return:
    """
    return isinstance(value, dict)


def is_tuple(value):
    """
    Check if an object is a tuple
    :param value:
    :return:
    """
    return isinstance(value, tuple)

def is_list_or_tuple(value):
    """
    Check if an object is a list or a tuple
    :param value:
    :return:
    """
    return isinstance(value, list)

def is_list_of_int(value):
    """
    Check if an object is a list of integers
    :param value:
    :return:
    """
    return isinstance(value, list) and all(isinstance(elem, int) for elem in value)


def is_list_of_float(value):
    """
    Check if an object is a list of floats
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(isinstance(elem, float) for elem in value)


def is_list_of_str_or_int(value):
    """
    Check if an object is a string or an integer
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(isinstance(elem, (int, str)) for elem in value)


def is_list_of_str_or_num(value):
    """
    Check if an object is string, integer or float
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(isinstance(elem, (str, int, float)) for elem in value)


def is_list_of_dask_dataframes(value):
    """
    Check if an object is a Spark DataFrame
    :param value:
    :return:
    """
    return isinstance(value, list) and all(isinstance(elem, DaskDataFrame) for elem in value)


def is_filepath(file_path):
    """
    Check if a value ia a valid file path
    :param file_path:
    :return:
    """
    # the file is there
    if os.path.exists(file_path):
        return True
    # the file does not exists but write privileges are given
    elif os.access(os.path.dirname(file_path), os.W_OK):
        return True
    # can not write there
    else:
        return False


def is_ip(value):
    """
    Check if a value is valid ip
    :param value:
    :return:
    """
    return re.match(regex_ip_compiled, value)


def is_list_of_str(value):
    """
    Check if all elements in a list are strings
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(isinstance(elem, str) for elem in value)


def is_list_of_numeric(value):
    """
    Check if all elements in a list are int or float
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(isinstance(elem, (int, float)) for elem in value)


def is_list_of_list(value):
    """
    Check if all elements in a list are tuples
    :param value:
    :return:
    """

    return bool(value) and isinstance(value, list) and all(isinstance(elem, list) for elem in value)


def is_list_of_tuples(value):
    """
    Check if all elements in a list are tuples
    :param value:
    :return:
    """

    return bool(value) and isinstance(value, list) and all(isinstance(elem, tuple) for elem in value)


def is_list_of_one_element(value):
    """
    Check if a var is a single element
    :param value:
    :return:
    """
    if is_list_value(value):
        return len(value) == 1


def is_dict_of_one_element(value):
    """
    Check if a var is a single element
    :param value:
    :return:
    """
    if is_dict(value):
        return len(value) == 1


def is_one_element(value):
    """
    Check if a var is a single element
    :param value:
    :return:
    """
    return isinstance(value, (str, int, float, bool))


def is_num_or_str(value):
    """
    Check if a var is numeric(int, float) or string
    :param value:
    :return:
    """
    return isinstance(value, (int, float, str))


def is_str_or_int(value):
    """
    Check if a var is a single element
    :param value:
    :return:
    """

    return isinstance(value, (str, int))


def is_numeric(value):
    """
    Check if a var is a single element
    :param value:
    :return:
    """
    return isinstance(value, (int, float))


def is_str(value):
    """
    Check if an object is a string
    :param value:
    :return:
    """
    # Seems 20% faster than
    return isinstance(value, str)
    # return True if type("str") == "str" else False


def is_object_value(value):
    """
    Check if an object is a string
    :param value:
    :return:
    """
    return isinstance(value, object)


def is_list_of_futures(value):
    """
    Check if an object is a list of strings
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(
        isinstance(elem, distributed.client.Future) for elem in value)


def is_future(value):
    """
    Check if an object is a list of strings
    :param value:
    :return:
    """
    return isinstance(value, distributed.client.Future)


def is_decimal(value):
    return fastnumbers.isfloat(value, allow_nan=True)


def is_int(value):
    """
    Check if an object is an integer
    :param value:
    :return:
    """
    return isinstance(value, int) and not isinstance(value, bool)


def is_int_like(value):
    """
    Check if an object is an integer
    :param value:
    :return:
    """
    return fastnumbers.isintlike(value)


def is_url(value):
    regex = re.compile(
        r'^(?:http|ftp|hdfs)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    return re.match(regex, value)


def is_float(value):
    """
    Check if an object is an integer
    :param value:
    :return:
    """
    return isinstance(value, float)


def is_bool_value(value):
    return isinstance(value, bool)


def is_datetime(value):
    """
    Check if an object is a datetime
    :param value:
    :return:
    """
    result = False
    if isinstance(value, datetime.datetime):
        result = True
    # else:
    #     result = is_datetime_str(str(value))

    return result


def is_binary(value):
    """
    Check if an object is a bytearray
    :param value:
    :return:
    """
    return isinstance(value, bytearray)


def is_date(value):
    """
    Check if an object is a date
    :param value:
    :return:
    """

    return isinstance(value, datetime.date)



from optimus.tests.base import TestBase

class TestCols(TestBase):
    dict = {"A": [-1, 2, 3, -10, -1.5, 50]}
    load = {}  # {"type": "csv", "path": "foo.csv"} | {"path": "foo.csv"} | "foo.csv"

    def test_rows(self):
        self.assertGreater(self.df.rows.count(), -1)

    def test_name(self):
        self.assertGreater(len(str(self.df.cols.names()[0])), 0)

    def test_abs(self):
        # self.create_dataframe uses the config of the test
        # expected_df = self.create_dataframe({"A": [1, 2, 3, 10, 1.5, 50]})

        # can compare an Optimus dataframe and a Python dictionary
        expected_df = {"A": [1, 2, 3, 10, 1.5, 50]}

        self.assertTrue(self.df.cols.abs().equals(expected_df))


class TestColsDask(TestCols):
    config = {"engine": "dask", "n_partitions": 1}


class TestColsDask2(TestCols):
    config = {"engine": "dask", "n_partitions": 2}


class TestColsDask3(TestCols):
    config = {"engine": "dask", "n_partitions": 3}



from abc import abstractmethod
from functools import reduce
from optimus.engines.base.basedataframe import BaseDataFrame
from optimus.infer import is_dict
from optimus.helpers.raiseit import RaiseIt
from optimus.helpers.core import val_to_list

from sklearn.preprocessing import MinMaxScaler, MaxAbsScaler, StandardScaler

from optimus.helpers.columns import parse_columns, name_col
from optimus.helpers.constants import Actions


class DataFrameBaseColumns():

    def _names(self):
        return list(self.root.data.columns)

    @abstractmethod
    def _pd(self):
        pass

    def append(self, dfs):
        """

        :param dfs:
        :return:
        """

        dfs = val_to_list(dfs)

        df = self.root
        
        dfd = df.data.reset_index(drop=True)

        dfds = []

        for _df in dfs:
            if is_dict(_df):
                _dfd = dfd
                for col, value in _df.items():
                    if isinstance(value, (BaseDataFrame,)):
                        value = value.data[value.cols.names()[0]]
                    _dfd[col] = value
                _df = _dfd[list(_df.keys())]
                del _dfd
            else:
                _df = _df.data
            dfds.append(_df)

        _dfd = self._pd.concat([_df.reset_index(drop=True) for _df in dfds], axis=1)

        for col in _dfd.columns:
            dfd[col] = _dfd[col]

        return self.root.new(dfd)

    def heatmap(self, col_x, col_y, bins_x=10, bins_y=10):
        pass

    def crosstab(self, col_x, col_y, output="dict"):
        """
        :param col_x:
        :param col_y:
        :param output:
        :return:
        """
        dfd = self.root.data

        result = self._pd.crosstab(dfd[col_x], dfd[col_y])

        if output == "dict":
            result = result.to_dict()
        elif output == "dataframe":
            result = self.root.new(result.reset_index())
        else:
            RaiseIt.value_error(output, ["dict", "dataframe"])

        return result

    def standard_scaler(self, input_cols="*", output_cols=None):
        df = self.root

        def _standard_scaler(_value):
            return StandardScaler().fit_transform(_value.values.reshape(-1, 1))

        return df.cols.apply(input_cols, func=_standard_scaler, output_cols=output_cols,
                             meta_action=Actions.STANDARD_SCALER.value)

    def max_abs_scaler(self, input_cols="*", output_cols=None):

        df = self.root

        def _max_abs_scaler(_value):
            return MaxAbsScaler().fit_transform(_value.values.reshape(-1, 1))

        return df.cols.apply(input_cols, func=_max_abs_scaler, output_cols=output_cols,
                             meta_action=Actions.MAX_ABS_SCALER.value)

    def min_max_scaler(self, input_cols, output_cols=None):
        # https://github.com/dask/dask/issues/2690

        df = self.root

        def _min_max_scaler(_value):
            return MinMaxScaler().fit_transform(_value.values.reshape(-1, 1))

        return df.cols.apply(input_cols, func=_min_max_scaler, output_cols=output_cols,
                             meta_action=Actions.MIN_MAX_SCALER.value)

    def replace_regex(self, input_cols, regex=None, value="", output_cols=None):
        """
        Use a Regex to replace values
        :param input_cols: '*', list of columns names or a single column name.
        :param output_cols:
        :param regex: values to look at to be replaced
        :param value: new value to replace the old one
        :return:
        """

        df = self.root

        def _replace_regex(_value, _regex, _replace):
            return _value.replace(_regex, _replace, regex=True)

        return df.cols.apply(input_cols, func=_replace_regex, args=(regex, value,), output_cols=output_cols,
                             filter_col_by_dtypes=df.constants.STRING_TYPES + df.constants.NUMERIC_TYPES)

    def reverse(self, input_cols, output_cols=None):
        def _reverse(value):
            return str(value)[::-1]

        df = self.root
        return df.cols.apply(input_cols, _reverse, func_return_type=str,
                             filter_col_by_dtypes=df.constants.STRING_TYPES,
                             output_cols=output_cols, mode="map", set_index=True)

    @staticmethod
    def astype(*args, **kwargs):
        pass

    @staticmethod
    def to_timestamp(input_cols, date_format=None, output_cols=None):
        pass

    def nest(self, input_cols, separator="", output_col=None, shape="string", drop=False):
        df = self.root

        dfd = df.data

        if output_col is None:
            output_col = name_col(input_cols)

        input_cols = parse_columns(df, input_cols)

        output_ordered_columns = df.cols.names()

        # cudfd do nor support apply or agg join for this operation
        if shape == "vector" or shape == "array":
            dfd = dfd.assign(**{output_col: dfd[input_cols].values.tolist()})

        elif shape == "string":
            dfds = [dfd[input_col].astype(str) for input_col in input_cols]
            dfd = dfd.assign(**{output_col: reduce((lambda x, y: x + separator + y), dfds)})

        if output_col not in output_ordered_columns:
            col_index = output_ordered_columns.index(input_cols[-1]) + 1
            output_ordered_columns[col_index:col_index] = [output_col]

        if drop is True:
            for input_col in input_cols:
                if input_col in output_ordered_columns and input_col != output_col:
                    output_ordered_columns.remove(input_col)

        return self.root.new(dfd).cols.select(output_ordered_columns)



# DataFrame = pd.DataFrame

import cudf

from optimus.engines.base.commons.functions import to_float_cudf, to_integer_cudf
from optimus.helpers.core import val_to_list
from optimus.engines.base.functions import Functions


class CUDFFunctions(Functions):

    def word_tokenize(self, series):
        pass

    def to_float(self, series):
        return to_float_cudf(series)

    def to_integer(self, series):
        return to_integer_cudf(series)

    def to_string(self, series):
        return series.astype(str)

    def to_boolean(self, series):
        return series.astype(bool)

    def count_zeros(self, series, *args):
        # Cudf can not handle null so we fill it with non zero values.
        non_zero_value = 1
        return (self.to_float(series).fillna(non_zero_value).values == 0).sum()

    def kurtosis(self, series):
        return self.to_float(series).kurt()

    def skew(self, series):
        return self.to_float(series).skew()

    def exp(self, series):
        return cudf.exp(self.to_float(series))

    def sqrt(self, series):
        return self.to_float(series).sqrt()

    def unique_values(self, series, *args):
        # Cudf can not handle null so we fill it with non zero values.
        return self.to_string(series).unique()

    # def mod(self, other):
    #     
    #     return cudf.mod(self.to_float(series), other)

    def radians(self, series):
        return cudf.radians(self.to_float(series))

    def degrees(self, series):
        return cudf.degrees(self.to_float(series))

    def ln(self, series):
        return self.to_float(series).log()

    def log(self, series, base=10):
        return cudf.log(self.to_float(series)) / cudf.log(base)

    def ceil(self, series):
        return self.to_float(series).ceil()

    def floor(self, series):
        return self.to_float(series).floor()

    def sin(self, series):
        return self.to_float(series).sin()

    def cos(self, series):
        return self.to_float(series).cos()

    def tan(self, series):
        return self.to_float(series).tan()

    def asin(self, series):
        return self.to_float(series).asin()

    def acos(self, series):
        return self.to_float(series).acos()

    def atan(self, series):
        return self.to_float(series).atan()

    def sinh(self, series):
        return 1 / 2 * (cudf.exp(series) - cudf.exp(-series))

    def cosh(self, series):
        return 1 / 2 * (cudf.exp(series) + cudf.exp(-series))

    def tanh(self, series):
        return self.sinh() / self.cosh()

    def asinh(self, series):
        return 1 / self.sinh()

    def acosh(self, series):
        return 1 / self.cosh()

    def atanh(self, series):
        return 1 / self.tanh()

    def cut(self, bins):
        raise NotImplementedError("Not implemented yet https://github.com/rapidsai/cudf/issues/5589")

    def replace_chars(self, series, search, replace_by):
        replace_by = val_to_list(replace_by)
        return self.to_string_accessor(series).replace(search, replace_by, regex=False)

    def normalize_chars(self, series):
        if not series.isnull().all():
            return self.to_string_accessor(series).normalize_characters()
        else:
            return series

    def date_format(self, series, current_format=None, output_format=None):

        # Some formats are no supported yet. https://github.com/rapidsai/cudf/issues/5991
        return cudf.to_datetime(series, format=current_format, errors="coerce").dt.strftime(output_format)

    def days_between(self, date_format=None):

        raise NotImplementedError("Not implemented yet see https://github.com/rapidsai/cudf/issues/1041")
        # return cudf.to_datetime(series).astype('str', format=date_format) - datetime.now().date()



class BaseSave:

    def file(self, path, *args, **kwargs):
        raise NotImplementedError("Not implemented yet")
    
    def csv(self, path, *args, **kwargs):
        raise NotImplementedError("Not implemented yet")
    
    def xml(self, path, *args, **kwargs):
        raise NotImplementedError("Not implemented yet")
    
    def json(self, path, *args, **kwargs):
        raise NotImplementedError("Not implemented yet")

    def excel(self, path, *args, **kwargs):
        raise NotImplementedError("Not implemented yet")

    def avro(self, path, *args, **kwargs):
        raise NotImplementedError("Not implemented yet")
    
    def parquet(self, path, *args, **kwargs):
        raise NotImplementedError("Not implemented yet")

    def orc(self, path, *args, **kwargs):
        raise NotImplementedError("Not implemented yet")

    def hdf5(self, path, *args, **kwargs):
        raise NotImplementedError("Not implemented yet")



from pyspark.ml.classification import LogisticRegression
from pyspark.ml import Pipeline
from pyspark.ml.image import ImageSchema
from sparkdl import DeepImageFeaturizer
from pyspark.ml.evaluation import MulticlassClassificationEvaluator
from sparkdl import DeepImagePredictor
from pyspark.sql.functions import udf
from pyspark.sql.types import ArrayType, StringType


class DL:
    @staticmethod
    def image_classifier_lr(df, input_col="image", output_col="features", model_name="InceptionV3"):
        featurizer = DeepImageFeaturizer(inputCol=input_col, outputCol=output_col, modelName=model_name)
        lr = LogisticRegression(maxIter=10, regParam=0.05, elasticNetParam=0.3, labelCol="label")
        p = Pipeline(stages=[featurizer, lr])
        p_model = p.fit(df)
        return p_model, p_model.transform(df)

    @staticmethod
    def evaluate_image_classifier(df, model):
        tested_df = model.transform(df)
        evaluator = MulticlassClassificationEvaluator(metricName="accuracy")
        return evaluator.evaluate(tested_df.select("prediction", "label"))

    @staticmethod
    def image_predictor(path, input_col="image", output_col="predicted_labels", model_name="InceptionV3",
                        decode_predictions=True, topK=10):
        image_df = ImageSchema.readImages(path)
        predictor = DeepImagePredictor(inputCol=input_col, outputCol=output_col, modelName=model_name,
                                       decodePredictions=decode_predictions, topK=topK)
        preds = predictor.transform(image_df)
        firstelement = udf(lambda v: (str(v[0][1]), float(v[0][2])), ArrayType(StringType()))
        return preds.select(firstelement('predicted_labels').alias("predicted_labels"))



import ibis

from optimus.engines.base.engine import BaseEngine
from optimus.engines.ibis.io.jdbc import JDBC
from optimus.optimus import Engine
from optimus._version import __version__


class IbisEngine(BaseEngine):
    __version__ = __version__

    def __init__(self, verbose=False, comm=None, *args, **kwargs):
        self.engine = Engine.IBIS.value

        self.verbose(verbose)

        self.client = ibis

    @staticmethod
    def connect(driver=None, host=None, database=None, user=None, password=None, port=None, schema="public",
                oracle_tns=None, oracle_service_name=None, oracle_sid=None, presto_catalog=None,
                cassandra_keyspace=None, cassandra_table=None, bigquery_project=None, bigquery_dataset=None):
        """
        Create the JDBC string connection
        :return: JDBC object
        """

        return JDBC(host, database, user, password, port, driver, schema, oracle_tns, oracle_service_name, oracle_sid,
                    presto_catalog, cassandra_keyspace, cassandra_table, bigquery_project, bigquery_dataset)



# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.4'
#       jupytext_version: 1.1.1
#   kernel_info:
#     name: python3
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # This notebook create the tests in python code. All this cells must be run to executed the tests

# %load_ext autoreload
# %autoreload 2

# + {"outputHidden": false, "inputHidden": false}
import sys
sys.path.append("../..")
# -

from optimus import Optimus
from optimus.helpers.test import Test

op = Optimus("cudf", n_workers=1, threads_per_worker=8, processes=False, memory_limit="3G", comm=True)

# +
import pandas as pd
from pyspark.sql.types import *
from datetime import date, datetime

cols = [
    "names",
    "height(ft)", 
    "function", 
    "rank", 
    "age",
    "weight(t)",
    "japanese name",
    "last position seen",
    "date arrival",
    "last date seen",
    "attributes",
    "Date Type",
    "timestamp",
    "Cybertronian",
    "function(binary)",
    "NullType",
    "Mixed"
]

   
        
rows = [
    ("Optim'us", -28, "Leader", 10, 4000000, 4.30, ["Inochi", "Convoy"], "19.442735,-99.201111", "1980/04/10",
     "2016/09/10", [8.5344, 4300.0], date(2016, 9, 10), datetime(2014, 6, 24), True, bytearray("Leader", "utf-8"),
     None, "male"),
    ("bumbl#ebéé  ", 17, "Espionage", 7, 5000000, 2.0, ["Bumble", "Goldback"], "10.642707,-71.612534", "1980/04/10",
     "2015/08/10", [5.334, 2000.0], date(2015, 8, 10), datetime(2014, 6, 24), True, bytearray("Espionage", "utf-8"),
     None, "1.1.1.1" ),
    ("ironhide&", 26, "Security", 7, 7000000, 4.0, ["Roadbuster"], "37.789563,-122.400356", "1980/04/10",
     "2014/07/10", [7.9248, 4000.0], date(2014, 6, 24), datetime(2014, 6, 24), True, bytearray("Security", "utf-8"),
     None,"http://hi-optimus.com"),
    ("Jazz", 13, "First Lieutenant", 8, 5000000, 1.80, ["Meister"], "33.670666,-117.841553", "1980/04/10",
     "2013/06/10", [3.9624, 1800.0], date(2013, 6, 24), datetime(2014, 6, 24), True,
     bytearray("First Lieutenant", "utf-8"), None,"5123456789123456"),
    ("Megatron", None, "None", 10, 5000000, 5.70, ["Megatron"], None, "1980/04/10", "2012/05/10", [None, 5700.0],
     date(2012, 5, 10), datetime(2014, 6, 24), True, bytearray("None", "utf-8"), None, "11529"),
    ("Metroplex_)^$", 300, "Battle Station", 8, 5000000, None, ["Metroflex"], None, "1980/04/10", "2011/04/10",
     [91.44, None], date(2011, 4, 10), datetime(2014, 6, 24), True, bytearray("Battle Station", "utf-8"), None, "3"),
    (None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, "optimus"),

]

source_df = pd.DataFrame(columns = cols, 
             data= rows)

from dask import dataframe as dd
source_df = dd.from_pandas(source_df, npartitions=1)


# -

source_df.cols.dtypes()

# +
cols = [('new_col', str),('new_col_2',str)]
rows = [(1,2),(2,3),(3,3),(4,3),(5,3),(6,3),(7,3)]

# df_col = dd.from_pandas(pd.DataFrame(columns =cols , data= [(1,2),(2,3),(3,3),(4,3),(5,3),(6,3),(7,3)], dtype=dict(zip(cols,[int, str]))), npartitions=1)
df_col = op.create.df(cols, rows)
print(df_col)                         
# -

# ### End Init Section

# # Test

# ## Columns Test

t = Test(op, source_df, "df_cols_dask", imports=["import numpy as np",
                                            "nan = np.nan",
                                            "import datetime",], path="df_cols_dask", final_path="..")

# +


def func(col_name, attrs):
    return F.col(col_name) * 2


numeric_col = "height(ft)"
numeric_col_B = "rank"
numeric_col_C = "rank"
string_col = "function"
date_col = "date arrival"
date_col_B = "last date seen"
new_col = "new col"
array_col = "attributes"
# -

print(numeric_col_B)
t.create(None, "cols.impute", None, "df", None, numeric_col_B)
t.run()

# +
# op.create.df()
# -

source_df.cols.select("rank")

t.create(None, "cols.hist", None, "json", None, ["height(ft)", numeric_col_B], 4)

t.create(source_df, "cols.string_to_index", None, "df", None, "rank")

# +
# FIX at creation time we los the metadata. Need to find a way to put it on the dataframe creation
# t.create(source_df, "cols.index_to_string", None, "df", None, "rank***STRING_TO_INDEX")
# -

t.run()

t.create(None, "cols.remove", None, "df", None, string_col, "i")

t.run()

t.create(None, "cols.remove", "list", "df", None, string_col, ["a", "i", "Es"])

t.create(None, "cols.remove", "list_output", "df", None, string_col, ["a", "i", "Es"], output_cols=string_col + "_new")

t.run()

t.create(None, "cols.min", None, "json", None, numeric_col)

source_df.cols.names()

t.create(None, "cols.min", "all_columns", "json", None, "*")
t.run()

t.create(None, "cols.max", None, "json", None, numeric_col)

t.create(None, "cols.max", "all_columns", "json", None, "*")
t.run()

source_df.cols.min("rank")

t.create(None, "cols.range", None, "json", None, numeric_col)

source_df.display()

t.create(None, "cols.range", "all_columns", "json", None, "*")

t.run()

t.create(None, "cols.median", None, "json", None, numeric_col)

t.create(None, "cols.median", "all_columns", "json", None, "*")

t.run()

t.create(None, "cols.percentile", None, "json", None, numeric_col, [0.05, 0.25], 1)

t.create(None, "cols.percentile", "all_columns", "json", None, "*", [0.05, 0.25], 1)

# ## MAD

# +

source_df.cols.mad(numeric_col, True)
# -

t.create(None, "cols.mad", None, "json", None, numeric_col)

t.create(None, "cols.mad", "all_columns", "json", None, "*")

t.run()

t.create(None, "cols.std", None, "json", None, numeric_col)

t.create(None, "cols.std", "all_columns", "json", None, "*")
t.run()

t.create(None, "cols.kurt", None, "json", None, numeric_col)
t.run()

t.create(None, "cols.kurt", "all_columns", "json", None, "*")
t.run()

t.create(None, "cols.mean", None, "json",None, numeric_col)

t.create(None, "cols.mean", "all_columns", "json", None, "*")
t.run()

t.create(None, "cols.skewness", None, "json", None, numeric_col)

t.create(None, "cols.skewness", "all_columns", "json", None, "*")
t.run()

t.create(None, "cols.sum", None, "json", None,numeric_col)

t.create(None, "cols.sum", "all_columns", "json", None, "*")
t.run()

t.create(None, "cols.variance", None, "json",None, numeric_col)

t.create(None, "cols.variance", "all_columns", "json", None, "*")
t.run()

source_df.table()

t.create(None, "cols.abs", None, "df", None, "weight(t)")
t.run()

t.create(None, "cols.abs", "all_columns", "df", None, "*")

source_df.table()

source_df.mode(dropna=True)

t.create(None, "cols.mode", None, "json", None, numeric_col)

# %%time
t.create(None, "cols.mode", "all_columns", "json", None, "*")

t.run()

t.create(None, "cols.count", None, "json")

# ## Count na

t.create(None, "cols.count_na", None, "json", None, numeric_col)

t.create(None, "cols.count_na", "all_columns", "json", None, "*")

t.run()

t.create(None, "cols.count_zeros", None, "json", None, numeric_col)

source_df.compute()

t.create(None, "cols.count_zeros", "all_columns", "json", None, "*")
t.run()

source_df.cols.names()

# ## Value counts

t.create(None, "cols.value_counts", None, "json", None, numeric_col)
t.run()

t.create(None, "cols.value_counts", "all_columns", "json", None, "*")
t.run()

t.create(None, "cols.count_uniques", None, "json", None, numeric_col)
t.run()

t.create(None, "cols.count_uniques", "all_columns", "json", None, "*")
t.run()

t.create(None, "cols.unique", None, "df", None, numeric_col)
t.run()

t.create(None, "cols.unique", "all_columns", "df", None, "*")
t.run()

t.create(None, "cols.add", None, "df", None, [numeric_col, numeric_col_B])

t.create(None, "cols.add", "all_columns", "df", None, "*"),

t.create(None, "cols.sub", None, "df", None,[numeric_col, numeric_col_B])

t.create(None, "cols.sub", "all_columns", "df",None, "*")

t.create(None, "cols.mul", None, "df", None,[numeric_col, numeric_col_B])

t.create(None, "cols.mul", "all_columns", "df", None, "*")

t.create(None, "cols.div", None, "df", None,[numeric_col, numeric_col_B])

t.create(None, "cols.div", "all_columns", "df", None,"*")

t.create(None, "cols.z_score", None, "df", None,numeric_col)

source_df.compute()

t.create(None, "cols.z_score", "all_columns", "df", None, "*")

t.create(None, "cols.iqr", None, "json", None, None, numeric_col)

t.create(None, "cols.iqr", "all_columns", "json", None, "*")

t.run()

source_df.display()

t.create(None, "cols.lower", None, "df", None, string_col)

t.create(None, "cols.lower", "all_columns", "df", None, "*")

t.create(None, "cols.upper", None, "df", None, string_col)

t.create(None, "cols.upper", "all_columns", "df", None, "*")

t.create(None, "cols.trim", None, "df", None, string_col)

t.create(None, "cols.trim", "all_columns", "df", None, "*")

print(string_col)
t.create(None, "cols.reverse", None, "df", None, string_col)

t.create(None, "cols.reverse", "all_columns", "df", None, "*")

source_df.compute()

print(string_col)
t.create(None, "cols.remove_accents", None, "df", None, string_col)

t.create(None, "cols.remove_accents", "all_columns", "df", None,string_col)

source_df.table()

t.create(None, "cols.remove_special_chars", None, "df", None, string_col)

t.create(None, "cols.remove_special_chars", "all_columns", "df", None, "*")
t.run()
# t.create(None, "cols.value_counts", None, "json", None, numeric_col)

t.create(None, "cols.remove_white_spaces", None, "df", None, string_col)

t.create(None, "cols.remove_white_spaces", "all_columns", "df", None,"*")

print(date_col)
t.create(None, "cols.date_format", None, "df", None,date_col, "%Y/%m/%d", "%d/%m/%Y")

t.run()

t.create(None, "cols.date_format", "all_columns", "df", None, [date_col, date_col_B], "%Y/%m/%d", "%d/%m/%Y")

# t.create(None, "cols.years_between", None, "df", date_col, "yyyy/MM/dd")
t.create(None, "cols.years_between", None, "df", None, date_col, "%Y/%m/%d")

# t.create(None, "cols.years_between", "multiple_columns", "df", [date_col, date_col_B], "yyyy/MM/dd")
t.create(None, "cols.years_between", "multiple_columns", "df", None,[date_col, date_col_B], "%Y/%m/%d")

t.run()

print(numeric_col_B)
t.create(None, "cols.impute", None, "df", None, numeric_col_B)

source_df.compute()

# %%time
t.create(None, "cols.impute", "multiple_columns", "df", None, "names","categorical")


t.run()

# ## Hist

t.create(None, "cols.hist", None, "json", None, ["height(ft)", numeric_col_B], 4)
t.run() 

t.create(None, "cols.hist", "all_columns", "json", None, "*", 4)

t.run()

t.create(None, "cols.frequency", None, "dict", None, numeric_col_B, 4)
t.run()

t.create(None, "cols.frequency", "all_columns", "dict", None, "*", 4)
t.run()

t.create(None, "cols.schema_dtype", None, "json", None, numeric_col_B)

# Problems with casting
# t.delete(None, "cols.schema_dtype", "all_columns", "json", "*")
t.run()

t.create(None, "cols.dtypes", None, "json", None, numeric_col_B)

t.run()

t.create(None, "cols.dtypes", "all_columns", "json",None, "*")

t.create(None, "cols.names", None, "json")

source_df.compute()

print(numeric_col_B)
# t.create(None, "cols.qcut", None, "df",None, numeric_col_B, 4)

# +
# t.create(None, "cols.qcut", "all_columns", "df", None,"*", 4)
# -

t.create(None, "cols.clip", None, "df", None,numeric_col_B, 3, 5)

source_df.dtypes

t.create(None, "cols.clip", "all_columns", "df", None,"*", 3, 5)

source_df["names"].str.replace(r'\d+', '').compute()

print(string_col)
t.create(None, "cols.remove_numbers", None, "df", None, string_col)

t.create(None, "cols.replace", "full", "df", None, string_col, ["First Lieutenant", "Battle"], "Match",
         search_by="full")

t.create(None, "cols.replace", "words", "df", None, string_col, ["Security", "Leader"], "Match", search_by="words")
t.run()

t.create(None, "cols.replace", "chars", "df", None, string_col, ["F", "E"], "Match", search_by="chars")

# +
# t.create(None, "cols.replace", "numeric", "df", None, "age", 5000000, 5, search_by="numeric")
# -

t.run()

# Assert is failing I can see why
t.create(None, "cols.replace", "all_columns", "df", None, "*", ["Jazz", "Leader"], "Match")
t.run()

# +
# Its necesary to save the function
# t.delete(None, "cols.apply_expr", None, "df", numeric_col_B, func)

# +
# Its necesary to save the function
# t.delete(None, "cols.apply_expr", "all_columns", "df", [numeric_col_B, numeric_col_C], func)

# +
# t.create(None, "cols.append", "number", "df", None, new_col, 1)
# -

df_col.to_json()

t.delete(None, "cols.append", "dataframes", "df", None, df_col)
t.run()

# +
# t.delete(None, "cols.append", "advance", "df", [("new_col_4", "test"),
#                                                ("new_col_5", df[numeric_col_B] * 2),
#                                                ("new_col_6", [1, 2, 3])
#                                                ]),
# -


t.create(None, "cols.rename", None, "df", None, numeric_col_B, numeric_col_B + "(old)")

t.create(None, "cols.rename", "list", "df",None,
         [numeric_col, numeric_col + "(tons)", numeric_col_B, numeric_col_B + "(old)"])

t.create(None, "cols.rename", "function", "df", None, str.upper)

print(numeric_col_B)
t.create(None, "cols.drop", None, "df", None,numeric_col_B)

t.create(None, "cols.cast", None, "df", None, string_col, "string")

t.create(None, "cols.cast", "all_columns", "df", None, "*", "str")

t.run()

# +
# Problems with precision
# t.delete(None, "cols.cast", "vector", "df", array_col, Vectors)
# -

source_df.cols.keep(numeric_col_B).display()

t.create(None, "cols.keep", None, "df", None, numeric_col_B)

t.create(None, "cols.move", "after", "df",None, numeric_col_B, "after", array_col)

t.create(None, "cols.move", "before", "df",None, numeric_col_B, "before", array_col)

t.create(None, "cols.move", "beginning", "df", None, numeric_col_B, "beginning")

t.create(None, "cols.move", "end", "df", None, numeric_col_B, "end")

t.create(None, "cols.select", None, "df", None, 0, numeric_col)

t.create(None, "cols.select", "regex", "df", None, "n.*", regex=True),

t.create(None, "cols.sort", None, "df")
t.run()

t.create(None, "cols.sort", "desc", "df", None, "desc")

t.create(None, "cols.sort", "asc", "df", None, "asc")

t.run()

t.create(None, "cols.fill_na", None, "df", None, numeric_col, "1")

# +
# t.create(None, "cols.fill_na", "array", "df", None, "japanese name", ["1", "2"])
# -

t.run()

# t.create(None, "cols.fill_na", "bool", "df", None, "Cybertronian", False)
t.create(None, "cols.fill_na", None, "df", None, "Cybertronian", False)

t.run()

# + {"jupyter": {"outputs_hidden": true}}
t.create(None, "cols.fill_na", "all_columns", "df", None,["names", "height(ft)", "function", "rank", "age"], "2")
# -

# ## Nest

t.create(None, "cols.nest", None, "df", None, [numeric_col, numeric_col_B], separator=" ", output_col=new_col)

# +
# t.create(None, "cols.nest", "mix", "df", [F.col(numeric_col_C), F.col(numeric_col_B)], "E", separator="--")
# -

df_na = source_df.cols.drop("NullType").rows.drop_na("*")

# +


t.create(df_na, "cols.nest", "vector_all_columns", "df", None, [numeric_col_C, numeric_col_B], shape="vector",
         output_col=new_col)
# -

t.create(df_na, "cols.nest", "vector", "df", None, [numeric_col_C, numeric_col_B], shape="vector", output_col=new_col)

t.create(None, "cols.nest", "array", "df", None, [numeric_col, numeric_col_B, numeric_col_C], shape="array",
         output_col=new_col)

t.run()

# +
# import pandas as pd
# from pyspark.sql.types import *
# from datetime import date, datetime

# cols = [
#     ("col 1", "str"),
#      ("col 2", "str"),
#      ("col 3", "str"),
# ]

# rows = [
#      ("male", "male", 1),
#         ("optimus", "bumblebee", 1),
#         ("3", "4.1", 1),
#         ("true", "False", 1),
#         ("[1,2,3,4]", "(1,2,3,4)", 1),
#         ("{1,2,3,4}", "{'key1' :1 , 'key2':2}", 1),
#         ("1.1.1.1", "123.123.123.123", 1),
#         ("http://hi-optimuse.com", "https://hi-bumblebee.com", 1),
#         ("optimus@cybertron.com", "bumblebee@cybertron.com", 1),
#         ("5123456789123456", "373655783158306", 1),
#         ("11529", "30345", 1),
#         ("04/10/1980", "04/10/1980", 1),
#         ("null", "Null", 1),
#         ("", "", 1),
#         (None, None, 1)

# ]

# source_df = pd.DataFrame(columns = cols, 
#              data= rows)

# # from dask import dataframe as dd

# # a=pd.DataFrame(source_df.to_dict())
# from dask import dataframe as dd
# source_df = dd.from_pandas(source_df, npartitions=1)
# -

mismatch = {"names": "dd/mm/yyyy", "height(ft)": r'^([0-2][0-9]|(3)[0-1])(\/)(((0)[0-9])|((1)[0-2]))(\/)\d{4}$',
            "function": "yyyy-mm-dd"}

t.create(None, "cols.count_mismatch", "int", "json", None, {"Mixed": "int"})
t.run()

t.create(None, "cols.count_mismatch", "decimal", "json", None, {"Mixed": "decimal"})
t.run()

t.create(None, "cols.count_mismatch", "email", "json", None, {"Mixed": "email"})

t.create(None, "cols.count_mismatch", "url", "json", None, {"Mixed": "url"})

t.create(None, "cols.count_mismatch", "ip", "json", None, {"Mixed": "ip"})

t.create(None, "cols.count_mismatch", "boolean", "json", None, {"Mixed": "boolean"})

t.create(None, "cols.count_mismatch", "credit_card", "json", None, {"Mixed": "credit_card_number"})

t.create(None, "cols.count_mismatch", "object", "json", None, {"Mixed": "object"})

t.create(None, "cols.count_mismatch", "array", "json", None, {"Mixed": "array"})

t.create(None, "cols.count_mismatch", "date", "json", None, {"Mixed": "date"})

t.create(None, "cols.count_mismatch", "zip_code", "json", None, {"Mixed": "zip_code"})

t.create(None, "cols.count_mismatch", "string", "json", None, {"Mixed": "string"})

t.run()

# ## Unnest String

source_df.cols.select(3).compute()

source_df.columns

t.create(None, "cols.unnest", "string_index", "df", None, date_col, "/", splits=3)

t.create(None, "cols.unnest", "string_multi_index", "df", None, date_col, "/", splits=3, index=[1, 2])

t.create(None, "cols.unnest", "string_infer_split", "df", None, date_col, "/")

t.create(None, "cols.unnest", "string_no_index", "df", None, date_col, "/", splits=3)

t.create(None, "cols.unnest", "string_output_columns", "df", None, date_col, "/", splits=3,
         output_cols=["year", "month", "day"])

t.create(None, "cols.unnest", "array_index", "df", None, array_col, index=1, mode="array")

t.create(None, "cols.unnest", "array_multi_index", "df", None, array_col, index=[0, 1], mode="array")

t.run()

t.create(None, "cols.unnest", "string_multi_colum_multi_index_multi_output", "df", None,
         ["date arrival", "last date seen"], "/", index=[[0, 1], [0, 1]],
         output_cols=[("year1", "month1"), ("year2", "month2")])

t.create(None, "cols.unnest", "string_multi_colum_multi_output", "df", None, ["date arrival", "last date seen"], "/",
         output_cols=[("year1", "month1"), ("year2", "month2")])

t.create(None, "cols.unnest", "array", "df", None,array_col, mode="array")

t.create(None, "cols.unnest", "array_all_columns", "df", None,array_col, mode="array")

t.create(None, "cols.is_na", "all_columns", "df", None, "*")

t.create(None, "cols.is_na", None, "df", None, numeric_col)

t.run()

# +

import numpy as np
# -

nan = np.nan
import datetime

# +
actual_df = op.load.json('https://raw.githubusercontent.com/hi-primus/optimus/master/examples/data/foo.json', multiline=True)

# expected_df = op.create.df(
#     [('billingId', LongType(), True), ('birth', StringType(), True), ('dummyCol', StringType(), True),
#      ('firstName', StringType(), True), ('id', LongType(), True), ('lastName', StringType(), True),
#      ('price', LongType(), True), ('product', StringType(), True)],
#     [(123, '1980/07/07', 'never', 'Luis', 1, 'Alvarez$$%!', 10, 'Cake')])

# # assert (expected_df.collect() == actual_df.collect())

# from deepdiff import DeepDiff  # For Deep Difference of 2 objects

# actual_df.table()
# expected_df.table()

# a1 = actual_df.to_json()
# e1 = expected_df.to_json()


# +
# ddiff = DeepDiff(a1, e1, ignore_order=False)
# print(ddiff)
# -

# # Rows Test

t = Test(op, df, "df_rows", imports=["import numpy as np",
                                     "nan = np.nan",
                                     "import datetime"])

rows = [
    ("Optim'us", 28, "Leader", 10, 5000000, 4.30, ["Inochi", "Convoy"], "19.442735,-99.201111", "1980/04/10",
     "2016/09/10", [8.5344, 4300.0], date(2016, 9, 10), datetime(2014, 6, 24), True, bytearray("Leader", "utf-8"),
     None)
]

df = op.load.csv("https://raw.githubusercontent.com/hi-primus/optimus/master/examples/data/foo.csv")

t = Test(op, source_df, "op_io_dask", imports=["import numpy as np",
                                          "nan = np.nan",
                                          "import datetime",], path="op_io_dask", final_path="..")

t.create(op, "load.csv", "local_csv", "df", None, "../../examples/data/foo.csv")
t.create(op, "load.json", "local_json", "df",None,"../../examples/data/foo.json", multiline=True)
# t.create(op, "load.parquet", "local_parquet", "df", None, "../../examples/data/foo.parquet")
t.create(op, "load.csv", "remote_csv", "df", None,
         "https://raw.githubusercontent.com/hi-primus/optimus/master/examples/data/foo.csv")

t.create(op, "load.json", "remote_json", "df", None,
         "https://raw.githubusercontent.com/hi-primus/optimus/master/examples/data/foo.json", multiline= True)

# +
# # !pip install pyarrow

# +
# t.create(op, "load.parquet", "remote_parquet", "df", None,
#          "https://raw.githubusercontent.com/hi-primus/optimus/master/examples/data/foo.parquet")

# +
from optimus.profiler.profiler import Profiler

p = Profiler()

print(p.run(source_df1, "japanese name"))
# -

# df_string = source_df.cols.cast("*","str")
t.create(source_df, "save.csv", None, None, None, "test.csv")

t.create(None, "save.json", None, None, None, "test.json")

# +
# t.create(None, "save.parquet", None, None, None, "test.parquet")
# -

t.run()

source_df.table()

# # Ouliers

t = Test(op, source_df, "df_outliers_dask", imports=["import numpy as np",
                                                "nan = np.nan",
                                                "import datetime"], path="df_outliers_dask",final_path="..")

# +
from pyspark.sql import functions as F


def func(col_name, attrs):
    return F.col(col_name) * 2


numeric_col = "height(ft)"
numeric_col_B = "rank"
numeric_col_C = "rank"
string_col = "function"
date_col = "date arrival"
date_col_B = "last date seen"
new_col = "new col"
array_col = "attributes"
# -

source_df.table()

# ## Tukey

t.create(None, "outliers.tukey", None, "df", "select", numeric_col)

source_df.outliers.tukey(numeric_col).drop().table()

t.create(None, "outliers.tukey", None, "df", "drop", numeric_col)

t.create(None, "outliers.tukey", None, "json", "whiskers", numeric_col)

t.create(None, "outliers.tukey", None, "json", "count", numeric_col)

t.create(None, "outliers.tukey", None, "json", "non_outliers_count", numeric_col)

t.create(None, "outliers.tukey", None, "json", "info", numeric_col)

t.run()

# ## Zscore

threshold = 0.5

t.create(None, "outliers.z_score", None, "df", "select", numeric_col, threshold)

source_df.outliers.z_score('height(ft)', 0.5).select()

t.create(None, "outliers.z_score", None, "df", "drop", numeric_col, threshold)

t.create(None, "outliers.z_score", None, "json", "count", numeric_col, threshold)

t.create(None, "outliers.z_score", None, "json", "non_outliers_count", numeric_col, threshold)

t.create(None, "outliers.z_score", None, "json", "info", numeric_col, threshold)

t.run()

source_df.cols.mad(numeric_col)

# ## Modified Zscore

threshold = 0.5
relative_error = 10000

t.create(None, "outliers.modified_z_score", None, "df", "select", numeric_col, threshold, relative_error)

t.create(None, "outliers.modified_z_score", None, "df", "drop", numeric_col, threshold, relative_error)

t.create(None, "outliers.modified_z_score", None, "json", "count", numeric_col, threshold, relative_error)

t.create(None, "outliers.modified_z_score", None, "json", "non_outliers_count", numeric_col, threshold, relative_error)

t.create(None, "outliers.modified_z_score", None, "json", "info", numeric_col, threshold, relative_error)

t.run()

# ## Mad

threshold = 0.5
relative_error = 10000

t.create(None, "outliers.mad", None, "df", "select", numeric_col, threshold, relative_error)

t.create(None, "outliers.mad", None, "df", "drop", numeric_col, threshold, relative_error)

t.create(None, "outliers.mad", None, "json", "count", numeric_col, threshold, relative_error)

t.create(None, "outliers.mad", None, "json", "non_outliers_count", numeric_col, threshold, relative_error)

t.create(None, "outliers.mad", None, "json", "info", numeric_col, threshold, relative_error)

t.run()

# ## Keycolision

source_df = op.load.csv("../../examples/data/random.csv", header=True, sep=";", dtype={'ADDRESS2': 'object',
       'COCINA': 'object',
       'CONDO': 'object',
       'ENGANCHE': 'object',
       'EXCEDENTE': 'float64',
       'FECHA_Entregado': 'object',
       'FECHA_Liberado': 'object',
       'FechaVencimientoApartado': 'object',
       'LT': 'object',
       'MZ': 'object',
       'NIVEL': 'object',
       'NOUNI': 'object',
       'OTRO2': 'object',
       'PAGINI': 'object',
       'cCodigoPostal': 'object'})

source_df.table()

# +
t = Test(op, source_df, "df_keycollision_dask", imports=["import numpy as np",
                                                    "nan = np.nan",
                                                    "import datetime",],
         path="df_keycollision_dask", final_path="..")


# -

from optimus.engines.pandas.ml import keycollision as keyCol

# +
# def func(values, ele_func, args):
#     return values.apply(ele_func, args)

# def ele_func(values, args):
#     return values

# args= ""
# source_df.map_partitions(func, ele_func, args)

# + {"outputHidden": false, "inputHidden": false}
t.create(keyCol, "fingerprint", None, "df", None, source_df, "STATE")
t.run()

# + {"outputHidden": false, "inputHidden": false}
t.create(keyCol, "fingerprint_cluster", None, "json", None, source_df, "STATE")
# -

t.run()

# + {"outputHidden": false, "inputHidden": false}
t.create(keyCol, "n_gram_fingerprint", None, "df", None, source_df, "STATE")

# + {"outputHidden": false, "inputHidden": false}
t.create(keyCol, "n_gram_fingerprint_cluster", None, "json", None, source_df, "STATE", 2)

# + {"outputHidden": false, "inputHidden": false}
t.run()
# -

# ## Distance cluster

source_df = op.read.csv("../../examples/data/random.csv", header=True, sep=";").limit(1000)

# + {"outputHidden": false, "inputHidden": false}
t = Test(op, source_df, "df_distance_cluster", imports=[\
                                                        "import numpy as np",
                                                        "nan = np.nan",
                                                        "import datetime",
                                                        "from pyspark.sql import functions as F",
                                                        "from optimus.ml import distancecluster as dc"],
         path="df_distance_cluster", final_path="..")

from optimus.ml import distancecluster as dc

# -

df.table()

# + {"outputHidden": false, "inputHidden": false}
t.create(dc, "levenshtein_cluster", None, 'dict', None, source_df, "STATE")
# -

t.run()

df_cancer = op.spark.read.csv('../data_cancer.csv', sep=',', header=True, inferSchema=True)

columns = ['diagnosis', 'radius_mean', 'texture_mean', 'perimeter_mean', 'area_mean', 'smoothness_mean',
           'compactness_mean', 'concavity_mean', 'concave points_mean', 'symmetry_mean',
           'fractal_dimension_mean']

df_model.table()

# ## Row

source_df = op.create.df([
    ("words", "str", True),
    ("num", "int", True),
    ("animals", "str", True),
    ("thing", StringType(), True),
    ("second", "int", True),
    ("filter", StringType(), True)
],
    [
        ("  I like     fish  ", 1, "dog dog", "housé", 5, "a"),
        ("    zombies", 2, "cat", "tv", 6, "b"),
        ("simpsons   cat lady", 2, "frog", "table", 7, "1"),
        (None, 3, "eagle", "glass", 8, "c"),
    ])

from optimus.engines.spark.audf import abstract_udf as audf

t = Test(op, source_df, "df_rows_dask", imports=["import numpy as np",
                                            "nan = np.nan",                                            
                                            "import datetime",], path="df_rows_dask", final_path="..")

row = [("this is a word", 2, "this is an animal",
        "this is a thing", 64, "this is a filter",)]

t.create(None, "rows.append", None, "df", None, row)

fil = (source_df["num"] == 1)

t.create(None, "rows.select", None, "df", None, fil)

t.create(None, "rows.select_by_dtypes", None, "df", None, "filter", "integer")

fil = (source_df["num"] == 2) | (source_df["second"] == 5)
print(str(fil))
# type(fil)

t.create(None, "rows.drop", None, "df", None, fil)

t.create(None, "rows.drop_by_dtypes", None, "df", None, "filter", "integer")


def func_data_type(value, attr):
    return value > 1


a = audf("num", func_data_type, "boolean")

t.create(None, "rows.drop", "audf", "df", None, a)

t.create(None, "rows.sort", None, "df", None, "num", "desc")

t.create(None, "rows.is_in", None, "df", None, "num", 2)

t.create(None, "rows.between", None, "df", None, "second", 6, 8)

t.create(None, "rows.between", "equal", "df", None, "second", 6, 8, equal=True)

t.create(None, "rows.between", "invert_equal", "df", None, "second", 6, 8, invert=True, equal=True)

t.create(None, "rows.between", "bounds", "df", None, "second", bounds=[(6, 7), (7, 8)], invert=True, equal=True)

t.run()




import logging


class Singleton(object):
    _instances = {}

    def __new__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__new__(cls, *args, **kwargs)
        return cls._instances[cls]


class Logger(Singleton):
    def __init__(self):
        self.logger = logging.getLogger('optimus')
        self.is_active = False

    def level(self, log_level):
        """
        Set the logging message level
        :param log_level:
        :return:
        """
        self.logger.setLevel(log_level)

    def print(self, *args, **kwargs):
        """
        Print a message
        :param self:
        :return:
        """
        if self.is_active is True:
            self.logger.info(*args, **kwargs)

    def active(self, activate):
        """
        Turn on and off the logging message
        :param activate:
        :return:
        """
        self.is_active = activate


logger = Logger()
logger.level(logging.INFO)


def level(log_level):
    logger.level(log_level)


def info(message):
    logger.print(message)


def active(active=None):
    logger.active(active)



import fastnumbers
from pyspark.ml.linalg import VectorUDT
from pyspark.sql import DataFrame as SparkDataFrame, functions as F
from pyspark.sql.types import StringType, IntegerType, FloatType, DoubleType, BooleanType, StructType, ArrayType, \
    LongType, DateType, ByteType, ShortType, TimestampType, BinaryType, NullType

from optimus.infer import is_bool_value, is_list_value, is_datetime, is_date, is_binary, is_str, is_bool_str, str_to_date, \
    is_list_str

SPARK_DTYPES_DICT = {"string": StringType, "int": IntegerType, "float": FloatType,
                     "double": DoubleType, "boolean": BooleanType, "struct": StructType, "array": ArrayType,
                     "bigint": LongType, "date": DateType, "byte": ByteType, "short": ShortType,
                     "datetime": TimestampType, "binary": BinaryType, "null": NullType, "vector": VectorUDT
                     }
SPARK_DTYPES_DICT_OBJECTS = \
    {"string": StringType(), "int": IntegerType(), "float": FloatType(),
     "double": DoubleType(), "boolean": BooleanType(), "struct": StructType(), "array": ArrayType(StringType()),
     "bigint": LongType(), "date": DateType(), "byte": ByteType(), "short": ShortType(),
     "datetime": TimestampType(), "binary": BinaryType(), "null": NullType()
     }
SPARK_DTYPES_TO_PROFILER = {"int": ["smallint", "tinyint", "bigint", "int"], "decimal": ["float", "double"],
                            "string": "string", "date": {"date", "timestamp"}, "boolean": "boolean", "binary": "binary",
                            "array": "array", "object": "object", "null": "null", "missing": "missing"}
PYSPARK_NUMERIC_TYPES = ["byte", "short", "big", "int", "double", "float"]
PYSPARK_NOT_ARRAY_TYPES = ["byte", "short", "big", "int", "double", "float", "string", "date", "bool"]
PYSPARK_STRING_TYPES = ["str"]
PYSPARK_ARRAY_TYPES = ["array"]
SPARK_SHORT_DTYPES = {"string": "string",
                      "str": "string",
                      "integer": "int",
                      "int": "int",
                      "bigint": "bigint",
                      "big": "bigint",
                      "long": "bigint",
                      "float": "float",
                      "double": "double",
                      "bool": "boolean",
                      "boolean": "boolean",
                      "struct": "struct",
                      "array": "array",
                      "date": "date",
                      "datetime": "datetime",
                      "byte": "byte",
                      "short": "short",
                      "binary": "binary",
                      "null": "null",
                      "vector": "vector",
                      "timestamp": "datetime"
                      }


def parse_spark_class_dtypes(value):
    """
    Get a pyspark data class from a string data type representation. for example 'StringType()' from 'string'
    :param value:
    :return:
    """
    if not isinstance(value, list):
        value = [value]

    try:
        data_type = [SPARK_DTYPES_DICT_OBJECTS[SPARK_SHORT_DTYPES[v]] for v in value]

    except (KeyError, TypeError):
        data_type = value

    if isinstance(data_type, list) and len(data_type) == 1:
        result = data_type[0]
    else:
        result = data_type

    return result


def to_spark(value):
    """
    Infer a Spark data type from a value
    :param value: value to be inferred
    :return: Spark data type
    """
    result = None
    if value is None:
        result = "null"

    elif is_bool_value(value):
        result = "bool"

    elif fastnumbers.isint(value):
        result = "int"

    elif fastnumbers.isfloat(value):
        result = "float"

    elif is_list_value(value):
        result = ArrayType(to_spark(value[0]))

    elif is_datetime(value):
        result = "datetime"

    elif is_date(value):
        result = "date"

    elif is_binary(value):
        result = "binary"

    elif is_str(value):
        if is_bool_str(value):
            result = "bool"
        elif is_datetime(value):
            result = "string"  # date
        elif is_list_str(value):
            result = "string"  # array
        else:
            result = "string"

    return parse_spark_class_dtypes(result)


def is_list_of_spark_dataframes(value):
    """
    Check if an object is a Spark DataFrame
    :param value:
    :return:
    """
    return bool(value) and isinstance(value, list) and all(isinstance(elem, SparkDataFrame) for elem in value)


def is_column(value):
    """
    Check if a object is a column
    :return:
    """
    return isinstance(value, F.Column)


from optimus.helpers.core import one_list_to_val


class RaiseIt:

    @staticmethod
    def type_error(var, data_types):
        """
        Raise a TypeError exception
        :param var:
        :param data_types: data types expected as strings
        :return:
        """

        from optimus.helpers.debug import get_var_name
        if len(data_types) == 1:
            divisor = ""
        elif len(data_types) == 2:
            divisor = " or "
        elif len(data_types) > 2:
            divisor = ", "

        _type = divisor.join(map(lambda x: "'" + x + "'", data_types))

        raise TypeError(
            "'{var_name}' must be of type {type}, received '{var_type}'".format(var_name=get_var_name(var), type=_type,
                                                                                var_type=type(var)))

    @staticmethod
    def length_error(var1: list, var2: (list, int)) -> Exception:
        """
        Raise a ValueError exception
        :param var1: variable to check for length
        :param var2: variable or integer to compare with var1
        :return:
        """
        from optimus.helpers.debug import get_var_name

        if isinstance(var2, int):
            length_var2 = str(var2)
        else:
            length_var2 = str(len(var2))

        raise ValueError("'{var2_name}' must be length '{var1_length}', received '{var2_length}'"
                         .format(var2_name=get_var_name(var2), var1_length=str(len(var1)), var2_length=length_var2))

    @staticmethod
    def not_ready_error(message):
        raise NotReady(message)

    @staticmethod
    def value_error(var=None, data_values=None, extra_text=""):
        """
        Raise a ValueError exception
        :param var:
        :param data_values: values accepted by the variable. str/list
        :param extra_text: Additional final info about the error
        :return:
        """
        from optimus.helpers.debug import get_var_name

        if not isinstance(data_values, list):
            data_values = [data_values]

        if len(data_values) == 1:
            divisor = ""
        elif len(data_values) == 2:
            divisor = " or "
        elif len(data_values) > 2:
            divisor = ", "

        raise ValueError("'{var_name}' must be {type}, received '{var_type}'. {extra_text}"
                         .format(var_name=get_var_name(var),
                                 type=divisor.join(map(
                                     lambda x: "'" + x + "'",
                                     data_values)), var_type=one_list_to_val(var), extra_text=extra_text))

    @staticmethod
    def type(cls, var):
        """
        Raise and exception of the type specified
        :param cls: Exception Class to be Raised
        :param var:
        :return:
        """

        from optimus.helpers.debug import get_var_name
        raise cls("'{var_name}' error".format(var_name=get_var_name(var), var_type=var))

    @staticmethod
    def message(cls, message):
        """
        Raise and exception of the type specified with a message
        :param cls: Exception Class to be Raised
        :param message:
        :return:
        """
        raise cls(message)


# define Python user-defined exceptions
class Error(Exception):
    """Base class for other exceptions"""
    pass


class NotReady(Error):
    pass



import pandas as pd

from optimus.engines.base.engine import BaseEngine
from optimus.engines.pandas.create import Create
from optimus.engines.pandas.io.extract import Extract
from optimus.engines.pandas.io.load import Load
from optimus.engines.pandas.pandas import Pandas
from optimus.helpers.exceptions import UnsupportedOperationError
from optimus._version import __version__

Pandas.instance = None


class PandasEngine(BaseEngine):
    __version__ = __version__

    def __init__(self, verbose=False, *args, **kwargs):
        self.extract = Extract()

        self.verbose(verbose)

        Pandas.instance = pd

        self.client = pd

    @property
    def constants(self):
        from optimus.engines.pandas.constants import Constants
        return Constants()

    @property
    def create(self):
        return Create(self)

    @property
    def load(self):
        return Load(self)

    @property
    def engine(self):
        return "pandas"

    def remote_run(self, callback, *args, **kwargs):
        if kwargs.get("client_timeout"):
            del kwargs["client_timeout"]

        callback(op=self, *args, **kwargs)

    def remote_submit(self, callback, *args, **kwargs):
        return self.submit(callback, op=self, *args, **kwargs)

    def submit(self, func, *args, priority=0, pure=False, **kwargs):
        import uuid
        def _func():
            return func(*args, **kwargs)

        return {"result": _func, "key": str(uuid.uuid4()), "status": "finished"}



from pyspark.sql.types import *
from optimus import Optimus
from optimus.helpers.json import json_enconding
from optimus.helpers.functions import deep_sort
import unittest
from pyspark.ml.linalg import Vectors, VectorUDT, DenseVector
import numpy as np
nan = np.nan
import datetime
from pyspark.sql import functions as F
from optimus.profiler.profiler import Profiler
null = None
true = True
p= Profiler()
op = Optimus(master='local')
source_df=op.create.df([('names', StringType(), True),('height(ft)', ShortType(), True),('function', StringType(), True),('rank', ByteType(), True),('age', IntegerType(), True),('weight(t)', FloatType(), True),('japanese name', ArrayType(StringType(),True), True),('last position seen', StringType(), True),('date arrival', StringType(), True),('last date seen', StringType(), True),('attributes', ArrayType(FloatType(),True), True),('Date Type', DateType(), True),('timestamp', TimestampType(), True),('Cybertronian', BooleanType(), True),('function(binary)', BinaryType(), True),('NullType', NullType(), True)], [('Optimus', -28, 'Leader', 10, 5000000, 4.300000190734863, ['Inochi', 'Convoy'], '19.442735,-99.201111', '1980/04/10', '2016/09/10', [8.53439998626709, 4300.0], datetime.date(2016, 9, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Leader'), None), ('bumbl#ebéé  ', 17, 'Espionage', 7, 5000000, 2.0, ['Bumble', 'Goldback'], '10.642707,-71.612534', '1980/04/10', '2015/08/10', [5.334000110626221, 2000.0], datetime.date(2015, 8, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Espionage'), None), ('ironhide&', 26, 'Security', 7, 5000000, 4.0, ['Roadbuster'], '37.789563,-122.400356', '1980/04/10', '2014/07/10', [7.924799919128418, 4000.0], datetime.date(2014, 6, 24), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Security'), None), ('Jazz', 13, 'First Lieutenant', 8, 5000000, 1.7999999523162842, ['Meister'], '33.670666,-117.841553', '1980/04/10', '2013/06/10', [3.962399959564209, 1800.0], datetime.date(2013, 6, 24), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'First Lieutenant'), None), ('Megatron', None, 'None', 10, 5000000, 5.699999809265137, ['Megatron'], None, '1980/04/10', '2012/05/10', [None, 5700.0], datetime.date(2012, 5, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'None'), None), ('Metroplex_)^$', 300, 'Battle Station', 8, 5000000, None, ['Metroflex'], None, '1980/04/10', '2011/04/10', [91.44000244140625, None], datetime.date(2011, 4, 10), datetime.datetime(2014, 6, 24, 0, 0), True, bytearray(b'Battle Station'), None), (None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)])
class Test_df_profiler(unittest.TestCase):
	maxDiff = None
	@staticmethod
	def test_columns_agg():
		actual_df =p.columns_agg(source_df,'*')
		actual_df =json_enconding(actual_df)
		expected_value =json_enconding({'names': {'count_uniques': 5, 'min': 'Jazz', 'max': 'ironhide&', 'count_na': 1, 'stddev': None, 'kurtosis': None, 'mean': None, 'skewness': None, 'sum': None, 'variance': None, 'zeros': 0}, 'height(ft)': {'count_uniques': 5, 'min': -28, 'max': 300, 'count_na': 2, 'stddev': 132.66612, 'kurtosis': 0.13863, 'mean': 65.6, 'skewness': 1.4049, 'sum': 328, 'variance': 17600.3, 'zeros': 0, 'percentile': {'0.75': 26, '0.95': 300, '0.05': -28, '0.25': 13, '0.5': 17}, 'hist': [{'count': 4.0, 'lower': -28.0, 'upper': 54.0}, {'count': 0.0, 'lower': 54.0, 'upper': 136.0}, {'count': 0.0, 'lower': 136.0, 'upper': 218.0}, {'count': 0.0, 'lower': 218.0, 'upper': 300.0}]}, 'function': {'count_uniques': 6, 'min': 'Battle Station', 'max': 'Security', 'count_na': 1, 'stddev': None, 'kurtosis': None, 'mean': None, 'skewness': None, 'sum': None, 'variance': None, 'zeros': 0}, 'rank': {'count_uniques': 3, 'min': 7, 'max': 10, 'count_na': 1, 'stddev': 1.36626, 'kurtosis': -1.5, 'mean': 8.33333, 'skewness': 0.3818, 'sum': 50, 'variance': 1.86667, 'zeros': 0, 'percentile': {'0.75': 10, '0.95': 10, '0.05': 7, '0.25': 7, '0.5': 8}, 'hist': [{'count': 4.0, 'lower': 7.0, 'upper': 8.5}, {'count': 0.0, 'lower': 8.5, 'upper': 10.0}]}, 'age': {'count_uniques': 1, 'min': 5000000, 'max': 5000000, 'count_na': 1, 'stddev': 0.0, 'kurtosis': nan, 'mean': 5000000.0, 'skewness': nan, 'sum': 30000000, 'variance': 0.0, 'zeros': 0, 'percentile': {'0.75': 5000000, '0.95': 5000000, '0.05': 5000000, '0.25': 5000000, '0.5': 5000000}, 'hist': [{'count': 6, 'lower': 5000000, 'upper': 5000001}]}, 'weight(t)': {'count_uniques': 5, 'min': 1.8, 'max': 5.7, 'count_na': 2, 'stddev': 1.64712, 'kurtosis': -1.43641, 'mean': 3.56, 'skewness': 0.06521, 'sum': 17.8, 'variance': 2.713, 'zeros': 0, 'percentile': {'0.75': 4.300000190734863, '0.95': 5.699999809265137, '0.05': 1.7999999523162842, '0.25': 2.0, '0.5': 4.0}, 'hist': [{'count': 1.0, 'lower': 1.8, 'upper': 2.78}, {'count': 0.0, 'lower': 2.78, 'upper': 3.75}, {'count': 2.0, 'lower': 3.75, 'upper': 4.73}, {'count': 1.0, 'lower': 4.73, 'upper': 5.7}]}, 'japanese name': {'count_uniques': 6, 'min': ['Bumble', 'Goldback'], 'max': ['Roadbuster'], 'count_na': 1}, 'last position seen': {'count_uniques': 4, 'min': '10.642707,-71.612534', 'max': '37.789563,-122.400356', 'count_na': 3, 'stddev': None, 'kurtosis': None, 'mean': None, 'skewness': None, 'sum': None, 'variance': None, 'zeros': 0}, 'date arrival': {'count_uniques': 1, 'min': '1980/04/10', 'max': '1980/04/10', 'count_na': 1, 'stddev': None, 'kurtosis': None, 'mean': None, 'skewness': None, 'sum': None, 'variance': None, 'zeros': 0}, 'last date seen': {'count_uniques': 6, 'min': '2011/04/10', 'max': '2016/09/10', 'count_na': 1, 'stddev': None, 'kurtosis': None, 'mean': None, 'skewness': None, 'sum': None, 'variance': None, 'zeros': 0}, 'attributes': {'count_uniques': 6, 'min': [None, 5700.0], 'max': [91.44000244140625, None], 'count_na': 1}, 'Date Type': {'count_uniques': 6, 'min': datetime.date(2011, 4, 10), 'max': datetime.date(2016, 9, 10), 'count_na': 1}, 'timestamp': {'count_uniques': 1, 'min': datetime.datetime(2014, 6, 24, 0, 0), 'max': datetime.datetime(2014, 6, 24, 0, 0), 'count_na': 1}, 'Cybertronian': {'count_uniques': 1, 'min': 1, 'max': 1, 'count_na': 1}, 'function(binary)': {'count_uniques': 6, 'min': bytearray(b'Battle Station'), 'max': bytearray(b'Security'), 'count_na': 1}, 'NullType': {'count_uniques': 0, 'min': None, 'max': None, 'count_na': 7}, 'p_count_na': 100.0, 'p_count_uniques': 0.0, 'range': 3.9000000000000004, 'median': 4.0, 'interquartile_range': 2.3000001907348633, 'coef_variation': 0.46267, 'mad': 1.7})
		assert(expected_value == actual_df)



import dask.array as da
from dask_ml import preprocessing

from optimus.engines.base.commons.functions import impute, string_to_index, index_to_string
from optimus.engines.base.columns import BaseColumns
from optimus.engines.base.dask.columns import DaskBaseColumns
from optimus.helpers.columns import parse_columns
import dask

from optimus.profiler.constants import MAX_BUCKETS


class Cols(DaskBaseColumns, BaseColumns):
    def __init__(self, df):
        super().__init__(df)

    def _series_to_pandas(self, series):
        return series.compute()

    def _names(self):
        return list(self.root.data.columns)

    def string_to_index(self, input_cols=None, output_cols=None, columns=None):
        le = preprocessing.LabelEncoder()
        return string_to_index(self, input_cols, output_cols, le)

    def index_to_string(self, input_cols=None, output_cols=None, columns=None):
        le = preprocessing.LabelEncoder()
        return index_to_string(self, input_cols, output_cols, le)

    def hist(self, columns="*", buckets=MAX_BUCKETS, compute=True):
        df = self.root
        columns = parse_columns(df, columns)

        result = {}
        for col_name in columns:
            # dfd = df[col_name].cols.to_float().data
            dfd = df.data[col_name].astype("float")

            if len(dfd) > 0:
                _count, _bins = dask.compute(da.histogram(dfd, bins=buckets, range=[dfd.min(), dfd.max()]))[0]
                result[col_name] = [
                    {"lower": float(_bins[i]), "upper": float(_bins[i + 1]), "count": _count[i]}
                    for i in range(buckets)]

        return {"hist": result}

    def impute(self, input_cols, data_type="continuous", strategy="mean", fill_value=None, output_cols=None):
        df = self.root
        return impute(df, input_cols, data_type=data_type, strategy=strategy, output_cols=None)



from abc import abstractmethod
from optimus.engines.base.meta import Meta
from optimus.infer import is_tuple
from optimus.helpers.types import DataFrameType, InternalDataFrameType
import pandas as pd


class BaseCreate:
    def __init__(self, root):
        self.root = root

    def _dictionary(self, dict):

        new_dict = {}
        
        for key, values in dict.items():
            if is_tuple(key):
                if len(key) == 3:
                    name, dtype, nulls = key
                elif len(key) == 2:
                    name, dtype = key
                    nulls = False
                dtype = self.root.constants.DTYPES_ALIAS.get(dtype, dtype)
            else:
                name = key
                dtype = None
                nulls = False

            new_dict[(name, dtype, nulls)] = values

        return new_dict

    
    def _dfd_from_dict(self, dict):
        return pd.DataFrame({ name: pd.Series(values, dtype=dtype) for (name, dtype, nulls), values in dict.items() })

    @abstractmethod
    def _df_from_dfd(self, dfd, *args, **kwargs):
        pass


    def dataframe(self, dict: dict=None, dfd: InternalDataFrameType=None, n_partitions: int=1, *args, **kwargs) -> DataFrameType:
        """
        Creates a dictionary using the form 
        {"Column name": ["value 1", "value 2"], ...} or {("Column name", "str", True): ["value 1", "value 2"]}
        Where the tuple keys uses the form (str, str, boolean) for (column name, data type, allow nulls)
        :param dict: A dictionary to construct the dataframe for
        :param dfd: A pandas dataframe, ignores dict when passed
        :return: BaseDataFrame
        """

        if dfd is None:
            if dict is None:
                dict = kwargs
                kwargs = {}
            dict = self._dictionary(dict)
            dfd = self._dfd_from_dict(dict)

        df = self._df_from_dfd(dfd, n_partitions=n_partitions, *args, **kwargs)
        df.meta = Meta.set(df.meta, value={"max_cell_length": df.cols.len("*").cols.max()})
        return df




from pyspark.sql import functions as F

from optimus.helpers.core import one_list_to_val
from optimus.helpers.functions import is_pyarrow_installed
from optimus.helpers.logger import logger
from optimus.helpers.parser import parse_python_dtypes
from optimus.helpers.raiseit import RaiseIt
from optimus.infer import is_tuple
from optimus.infer_spark import parse_spark_class_dtypes, is_column


def abstract_udf(col, func, func_return_type=None, args=None, func_type=None):
    """
    Abstract User defined functions. This is a helper function to create udf, pandas udf or a Column Exp
    :param col: Column to created or transformed
    :param func: Function to be applied to the data
    :param args: If required attributes to be passed to the function
    :param func_return_type: Required by UDF and Pandas UDF.
    :param func_type: pandas_udf or udf. The function is going to try to use pandas_udf if func_type is not defined
    :return: A function, UDF or Pandas UDF
    """

    if func_return_type is None:
        func_type = "column_expr"
    # By default is going to try to use pandas UDF
    elif func_type is None and is_pyarrow_installed() is True:
        func_type = "pandas_udf"

    types = ["column_expr", "udf", "pandas_udf"]
    if func_type not in types:
        RaiseIt.value_error(func_type, types)

    # It handle if func param is a plain expression or a function returning and expression
    def func_col_exp(col_name, attr):
        return func

    if is_column(func):
        _func = func_col_exp
    else:
        _func = func
    # print(func_type)
    logger.print(
        "Using '{func_type}' to process column '{column}' with function {func_name}".format(func_type=func_type,
                                                                                            column=col,
                                                                                            func_name=_func.__name__))

    df_func = func_factory(func_type, func_return_type)
    if not is_tuple(args):
        args = (args,)

    # print("-----------------df_func(_func, args)(col)", df_func(_func, args)(col))
    return df_func(_func, args)(col)


def func_factory(func_type=None, func_return_type=None):
    """
    Return column express, udf or pandas udf function.
    :param func_type: Type of function udf or pandas udf
    :param func_return_type:
    :return:
    """

    func_return_type = parse_spark_class_dtypes(func_return_type)

    def pandas_udf_func(func=None, args=None):
        # TODO: Get the column type, so is not necessary to pass the return type as param.

        # Apply the function over the whole series
        def apply_to_series(value, args):
            if args is None or args == (None,):
                return value.apply(func)
            else:
                return value.apply(func, args=args)

        def to_serie(value):
            return apply_to_series(value, args)

        return F.pandas_udf(to_serie, func_return_type)

    def udf_func(func, args):
        if args is None or args == (None,):
            return F.udf(lambda value: func(value), func_return_type)
        else:
            return F.udf(lambda value: func(value, *args), func_return_type)

    def expression_func(func, args):
        def inner(col_name):
            if args is None or args == (None,):
                return func(col_name)
            else:
                return func(col_name, *args)

        return inner

    if func_type is "pandas_udf":
        return pandas_udf_func

    elif func_type is "udf":
        return udf_func

    elif func_type is "column_expr":
        return expression_func


def filter_row_by_data_type_audf(col_name, data_type):
    """
    Filter a column using a Spark data type as reference
    :param col_name:
    :param data_type:
    :return:
    """

    data_type = parse_python_dtypes(data_type)
    return abstract_udf(col_name, filter_row_by_data_type, "boolean", data_type)


def filter_row_by_data_type(col_name, data_type=None, get_type=False):
    """
    A Pandas UDF function that returns bool if the value match with the data_type param passed to the function.
    Also can return the data type
    :param col_name: Column to be process
    :param data_type: The data_type to be compared with
    :param get_type: Value to be returned as string or boolean
    :return: True or False
    """

    if data_type is not None:
        data_type = parse_python_dtypes(data_type)

    def pandas_udf_func(v):

        return v.apply(Infer.func, args=(data_type, get_type))

    if get_type is True:
        return_data_type = "string"
    else:
        return_data_type = "boolean"

    col_name = one_list_to_val(col_name)
    return F.pandas_udf(pandas_udf_func, return_data_type)(col_name)



import numpy as np
from pyspark.sql.types import *

from optimus import Optimus

nan = np.nan
import datetime

op = Optimus(master='local')
source_df = op.create.df(
    [('names', StringType(), True), ('height(ft)', ShortType(), True), ('function', StringType(), True),
     ('rank', ByteType(), True), ('age', IntegerType(), True), ('weight(t)', FloatType(), True),
     ('japanese name', ArrayType(StringType(), True), True), ('last position seen', StringType(), True),
     ('date arrival', StringType(), True), ('last date seen', StringType(), True),
     ('attributes', ArrayType(FloatType(), True), True), ('Date Type', DateType(), True),
     ('timestamp', TimestampType(), True), ('Cybertronian', BooleanType(), True),
     ('function(binary)', BinaryType(), True), ('NullType', NullType(), True)], [("Optim'us", 28, 'Leader', 10, 5000000,
                                                                                  4.300000190734863,
                                                                                  ['Inochi', 'Convoy'],
                                                                                  '19.442735,-99.201111', '1980/04/10',
                                                                                  '2016/09/10',
                                                                                  [8.53439998626709, 4300.0],
                                                                                  datetime.date(2016, 9, 10),
                                                                                  datetime.datetime(2014, 6, 24, 0, 0),
                                                                                  True, bytearray(b'Leader'), None), (
                                                                                     'bumbl#ebéé  ', 17, 'Espionage', 7,
                                                                                     5000000, 2.0,
                                                                                     ['Bumble', 'Goldback'],
                                                                                     '10.642707,-71.612534',
                                                                                     '1980/04/10',
                                                                                     '2015/08/10',
                                                                                     [5.334000110626221, 2000.0],
                                                                                     datetime.date(2015, 8, 10),
                                                                                     datetime.datetime(2014, 6, 24, 0,
                                                                                                       0),
                                                                                     True, bytearray(b'Espionage'),
                                                                                     None), (
                                                                                     'ironhide&', 26, 'Security', 7,
                                                                                     5000000, 4.0, ['Roadbuster'],
                                                                                     '37.789563,-122.400356',
                                                                                     '1980/04/10',
                                                                                     '2014/07/10',
                                                                                     [7.924799919128418, 4000.0],
                                                                                     datetime.date(2014, 6, 24),
                                                                                     datetime.datetime(2014, 6, 24, 0,
                                                                                                       0),
                                                                                     True, bytearray(b'Security'),
                                                                                     None), (
                                                                                     'Jazz', 13, 'First Lieutenant', 8,
                                                                                     5000000, 1.7999999523162842,
                                                                                     ['Meister'],
                                                                                     '33.670666,-117.841553',
                                                                                     '1980/04/10', '2013/06/10',
                                                                                     [3.962399959564209, 1800.0],
                                                                                     datetime.date(2013, 6, 24),
                                                                                     datetime.datetime(2014, 6, 24, 0,
                                                                                                       0),
                                                                                     True,
                                                                                     bytearray(b'First Lieutenant'),
                                                                                     None), (
                                                                                     'Megatron', None, 'None', 10,
                                                                                     5000000,
                                                                                     5.699999809265137, ['Megatron'],
                                                                                     None,
                                                                                     '1980/04/10', '2012/05/10',
                                                                                     [None, 5700.0],
                                                                                     datetime.date(2012, 5, 10),
                                                                                     datetime.datetime(2014, 6, 24, 0,
                                                                                                       0),
                                                                                     True, bytearray(b'None'), None), (
                                                                                     'Metroplex_)^$', 300,
                                                                                     'Battle Station',
                                                                                     8, 5000000, None, ['Metroflex'],
                                                                                     None,
                                                                                     '1980/04/10', '2011/04/10',
                                                                                     [91.44000244140625, None],
                                                                                     datetime.date(2011, 4, 10),
                                                                                     datetime.datetime(2014, 6, 24, 0,
                                                                                                       0),
                                                                                     True, bytearray(b'Battle Station'),
                                                                                     None)])
meta_value = {"key": "value"}


class Test_meta(object):

    @staticmethod
    def test_cols_set_meta():
        actual_df = source_df.cols.set_meta("height(ft)", value=meta_value)
        actual_json = actual_df._jdf.schema().json()
        expected_json = '{"type":"struct","fields":[{"name":"names","type":"string","nullable":true,"metadata":{}},{"name":"height(ft)","type":"short","nullable":true,"metadata":{"key":"value"}},{"name":"function","type":"string","nullable":true,"metadata":{}},{"name":"rank","type":"byte","nullable":true,"metadata":{}},{"name":"age","type":"integer","nullable":true,"metadata":{}},{"name":"weight(t)","type":"float","nullable":true,"metadata":{}},{"name":"japanese name","type":{"type":"array","elementType":"string","containsNull":true},"nullable":true,"metadata":{}},{"name":"last position seen","type":"string","nullable":true,"metadata":{}},{"name":"date arrival","type":"string","nullable":true,"metadata":{}},{"name":"last date seen","type":"string","nullable":true,"metadata":{}},{"name":"attributes","type":{"type":"array","elementType":"float","containsNull":true},"nullable":true,"metadata":{}},{"name":"Date Type","type":"date","nullable":true,"metadata":{}},{"name":"timestamp","type":"timestamp","nullable":true,"metadata":{}},{"name":"Cybertronian","type":"boolean","nullable":true,"metadata":{}},{"name":"function(binary)","type":"binary","nullable":true,"metadata":{}},{"name":"NullType","type":"null","nullable":true,"metadata":{}}]}'
        assert (actual_json == expected_json)

    @staticmethod
    def test_cols_get_meta():
        actual_df = source_df.cols.set_meta("height(ft)", value=meta_value)
        actual_value = actual_df.cols.get_meta("height(ft)")
        expected_value = {}
        assert (actual_value == expected_value)

    @staticmethod
    def test_df_set_meta():
        actual_df = source_df.meta.set(value=meta_value)
        actual_json = actual_df.schema[-1].metadata
        expected_json = meta_value
        assert (actual_json == expected_json)

    @staticmethod
    def test_df_get_meta():
        source_df.schema[-1].metadata = meta_value
        actual_value = source_df.meta.get()
        expected_value = meta_value
        assert (actual_value == expected_value)



import math

import simplejson as json

from optimus.helpers.constants import ProfilerDataTypes, CONFIDENCE_LEVEL_CONSTANT, PROFILER_COLUMN_TYPES
from optimus.helpers.json import json_converter


def fill_missing_var_types(var_types, dtypes):
    """
    Fill missing data types with 0
    :param var_types:
    :param dtypes:
    :return:
    """
    for k, v in var_types.items():
        if dtypes[k] == ProfilerDataTypes.STRING.value:
            for label in ProfilerDataTypes:
                if label.value not in v:
                    var_types[k][label.value] = 0
                else:
                    var_types[k] = v
        else:
            var_types[k] = v
    return var_types


def fill_missing_col_types(col_types):
    """
    Fill missing col types with 0
    :param col_types:
    :return:
    """
    for label in PROFILER_COLUMN_TYPES:
        if label not in col_types:
            col_types[label] = 0
    return col_types


def write_json(data, path):
    """
    Write a json file with the profiler result
    :param data:
    :param path:
    :return:
    """
    try:
        with open(path, 'w', encoding='utf-8') as outfile:
            json.dump(data, outfile, indent=4, ensure_ascii=False, default=json_converter)
    except IOError:
        pass


def write_html(data, path):
    """
    Write a json file with the profiler result
    :param data:
    :param path:
    :return:
    """

    try:
        with open(path, 'w', encoding='utf-8') as outfile:
            outfile.write(data)
    except IOError:
        pass


def sample_size(population_size, confidence_level, confidence_interval):
    """
    Get a sample number of the whole population
    :param population_size:
    :param confidence_level:
    :param confidence_interval:
    :return:
    """
    z = 0.0
    p = 0.5
    e = confidence_interval / 100.0
    n = population_size

    # Loop through supported confidence levels and find the num sdd deviations for that confidence level
    for i in CONFIDENCE_LEVEL_CONSTANT:
        if i[0] == confidence_level:
            z = i[1]

    if z == 0.0:
        return -1

    # Calculate sample size
    n_0 = ((z ** 2) * p * (1 - p)) / (e ** 2)

    # Adjust sample size fo finite population
    n = n_0 / (1 + ((n_0 - 1) / float(n)))

    return int(math.ceil(n))  # sample size
